--
-- DrivableHandtool
--
-- Author: Tobias Fricke
-- Description: Specialization for scythe mower functionality
-- Name: DrivableHandtool
-- Hide: yes
--
-- Copyright (c) Creative Mesh, 2024

---@class DrivableHandtool
DrivableHandtool = {}

DrivableHandtool.MOD_NAME = g_currentModName
DrivableHandtool.MOD_DIRECTORY = g_currentModDirectory
DrivableHandtool.MAX_FORWARD_SPEED = 3.9 -- km/h
DrivableHandtool.MAX_BACKWARD_SPEED = 2.9 -- km/h

function DrivableHandtool.initSpecialization()
    local schema = Vehicle.xmlSchema

    schema:setXMLSpecializationType("DrivableHandtool")
    schema:setXMLSpecializationType()
end

function DrivableHandtool.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(CCTDrivable, specializations)
end

function DrivableHandtool.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "resetInputs", DrivableHandtool.resetInputs)
    SpecializationUtil.registerFunction(vehicleType, "updateKinematic", DrivableHandtool.updateKinematic)
    SpecializationUtil.registerFunction(vehicleType, "getPosition", DrivableHandtool.getPosition)
    SpecializationUtil.registerFunction(vehicleType, "getRotation", DrivableHandtool.getRotation)
    SpecializationUtil.registerFunction(vehicleType, "getAccelerationAxis", DrivableHandtool.getAccelerationAxis)
    SpecializationUtil.registerFunction(vehicleType, "getCruiseControlDisplayInfo", DrivableHandtool.getCruiseControlDisplayInfo)
end

function DrivableHandtool.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", DrivableHandtool)
    SpecializationUtil.registerEventListener(vehicleType, "onLoadFinished", DrivableHandtool)
    SpecializationUtil.registerEventListener(vehicleType, "onReadUpdateStream", DrivableHandtool)
    SpecializationUtil.registerEventListener(vehicleType, "onWriteUpdateStream", DrivableHandtool)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdate", DrivableHandtool)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdateInterpolation", DrivableHandtool)
    SpecializationUtil.registerEventListener(vehicleType, "onLeaveVehicle", DrivableHandtool)
    SpecializationUtil.registerEventListener(vehicleType, "onRegisterActionEvents", DrivableHandtool)
end

function DrivableHandtool.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "setWorldPosition", DrivableHandtool.setWorldPosition)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "setWorldPositionQuaternion", DrivableHandtool.setWorldPositionQuaternion)
end

function DrivableHandtool:onLoad(savegame)
    self.spec_drivableHandtool = self[string.format("spec_%s.drivableHandtool", DrivableHandtool.MOD_NAME)]
    local spec = self.spec_drivableHandtool

    spec.currentDirtScale = 0
    spec.fadeDuration = 400
    spec.currentTurnAngle = 0

    spec.maxAcceleration = 5
    spec.maxDeceleration = 10
    spec.gravity = -9.81

    spec.inputValues = {
        moveForward = 0,
        moveRight = 0,
        lookRight = 0,
    }

    spec.axisForwardSend = 0
    spec.axisForward = 0
    spec.axisSideSend = 0
    spec.axisSide = 0
    spec.axisLookRightSend = 0
    spec.axisLookRight = 0

    self:resetInputs()

    if self.isServer then
        spec.interpolatorTurnAngle = InterpolatorAngle.new(0)
        self.networkTimeInterpolator.maxInterpolationAlpha = 1.2
    end

    spec.dirtyFlag = self:getNextDirtyFlag()

    -- reset movement collisionMask
	local spec_cctdrivable = self.spec_cctdrivable
	spec_cctdrivable.movementCollisionMask = 18
end

function DrivableHandtool:onLoadFinished()
    self:raiseActive()
end

---Called on read stream on update
---@param streamId number stream id
---@param timestamp number timestamp
---@param connection number connection id
function DrivableHandtool:onReadUpdateStream(streamId, timestamp, connection)
    local spec = self.spec_drivableHandtool

    if streamReadBool(streamId) then
        spec.axisForward = streamReadUIntN(streamId, 10) / 1023 * 2 - 1
        if math.abs(spec.axisForward) < 0.00099 then
            spec.axisForward = 0
        end

        spec.axisSide = streamReadUIntN(streamId, 10) / 1023 * 2 - 1
        if math.abs(spec.axisSide) < 0.00099 then
            spec.axisSide = 0
        end

        spec.axisLookRight = streamReadUIntN(streamId, 10) / 1023 * 2 - 1
        if math.abs(spec.axisLookRight) < 0.00099 then
            spec.axisLookRight = 0
        end

        self:raiseActive()
    end
end

---Called on write stream on update
---@param streamId number stream id
---@param connection number connection id
---@param dirtyMask number dirty mask
function DrivableHandtool:onWriteUpdateStream(streamId, connection, dirtyMask)
    local spec = self.spec_drivableHandtool

    if streamWriteBool(streamId, bitAND(dirtyMask, spec.dirtyFlag) ~= 0) then
        local axisForward = (spec.axisForward + 1) / 2 * 1023
        streamWriteUIntN(streamId, axisForward, 10)

        local axisSide = (spec.axisSide + 1) / 2 * 1023
        streamWriteUIntN(streamId, axisSide, 10)

        local axisLookRight = (spec.axisLookRight + 1) / 2 * 1023
        streamWriteUIntN(streamId, axisLookRight, 10)
    end
end

---Called on update
---@param dt number delta time since last update
function DrivableHandtool:onUpdate(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    local spec = self.spec_drivableHandtool

    if self:getIsEntered() and self.isClient then
        spec.axisForward = spec.inputValues.moveForward
        spec.axisSide = spec.inputValues.moveRight
        spec.axisLookRight = spec.inputValues.lookRight

        self:resetInputs()

        if spec.axisForward ~= spec.axisForwardSend or spec.axisSide ~= spec.axisSideSend or spec.axisLookRight ~= spec.axisLookRightSend then
            spec.axisForwardSend = spec.axisForward
            spec.axisSideSend = spec.axisSide
            spec.axisLookRightSend = spec.axisLookRight

            self:raiseDirtyFlags(spec.dirtyFlag)
        end
    end
end

---Called on update interpolation
---@param dt number delta time since last update
function DrivableHandtool:onUpdateInterpolation(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    local spec = self.spec_drivableHandtool

    if not self.isServer then
        return
    end

    self:updateKinematic(dt)

    if self:getIsEntered() and not self.isClient then
        self:resetInputs()
    end

    local component = self.components[1]
    local x, y, z = self:getCCTWorldTranslation()

    component.networkInterpolators.position:setTargetPosition(x, y, z)
    spec.interpolatorTurnAngle:setTargetAngle(spec.currentTurnAngle)

    local phaseDuration = dt + 30
    self.networkTimeInterpolator:startNewPhase(phaseDuration)
    self.networkTimeInterpolator:update(dt)

    x, y, z = component.networkInterpolators.position:getInterpolatedValues(self.networkTimeInterpolator.interpolationAlpha)
    setTranslation(self.rootNode, x, y, z)

    local turnAngle = spec.interpolatorTurnAngle:getInterpolatedValue(self.networkTimeInterpolator.interpolationAlpha)
    local dirX = math.sin(turnAngle)
    local dirZ = math.cos(turnAngle)

    setDirection(self.rootNode, dirX, 0, dirZ, 0, 1, 0)
end

---Called on leaving vehicle
function DrivableHandtool:onLeaveVehicle()
    local spec = self.spec_drivableHandtool

    spec.axisForward = 0
    spec.axisSide = 0
    spec.axisLookRight = 0

    if self.isClient then
        self:resetInputs()
        self:raiseActive()
    end
end

---Returns acceleration axis
---@return integer accelerationAxis
function DrivableHandtool:getAccelerationAxis()
    if self.movingDirection == 0 then
        return 0
    end

    return self.movingDirection > -1 and 1 or -1
end

---Returns cruise control display info
---@return integer cruiseControlSpeed
---@return boolean isActive
function DrivableHandtool:getCruiseControlDisplayInfo()
    return 0, false
end

---Reset input values
function DrivableHandtool:resetInputs()
    local spec = self.spec_drivableHandtool

    spec.inputValues.moveForward = 0
    spec.inputValues.moveRight = 0
    spec.inputValues.lookRight = 0
end

---Update kinematics
---@param dt number delta time since last update
function DrivableHandtool:updateKinematic(dt)
    --only server side
    local spec = self.spec_drivableHandtool

    if spec.axisLookRight ~= 0 then
        spec.currentTurnAngle = spec.currentTurnAngle - spec.axisLookRight
    end

    local movementFactor = g_rebelMoon ~= nil and g_rebelMoon:getMovementFactor() or 1
    local inputX = spec.axisSide * movementFactor
    local inputZ = spec.axisForward * movementFactor
    if inputX == 0 and inputZ == 0 then
        return
    end

    local dtInSec = dt * 0.001
    local normInputX, normInputZ = MathUtil.vector2Normalize(inputX, inputZ)
    local currentWorldDirX, _, currentWorldDirZ = localDirectionToWorld(self.rootNode, normInputX, 0, normInputZ)
    currentWorldDirX, currentWorldDirZ = MathUtil.vector2Normalize(currentWorldDirX, currentWorldDirZ)

    local speedLimit, _ = self:getSpeedLimit(true)
    speedLimit = math.min(speedLimit, DrivableHandtool.MAX_FORWARD_SPEED)
    if self.movingDirection < 0 then
        speedLimit = math.min(speedLimit, DrivableHandtool.MAX_BACKWARD_SPEED)
    end

    local movement = speedLimit * dtInSec
    local movementX = movement * currentWorldDirX
    local movementY = spec.gravity * dtInSec
    local movementZ = movement * currentWorldDirZ

    self:moveCCT(movementX, movementY, movementZ)
end

------------------
---ActionEvents---
------------------

---Called on register action events
function DrivableHandtool:onRegisterActionEvents(isActiveForInput, isActiveForInputIgnoreSelection)
    if self.isClient then
        local spec = self.spec_drivableHandtool

        self:clearActionEventsTable(spec.actionEvents)

        if isActiveForInputIgnoreSelection then
            local _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.AXIS_MOVE_FORWARD_PLAYER, self, DrivableHandtool.actionEventMoveForwardBackward, false, false, true, true, nil)
            g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_VERY_HIGH)
            g_inputBinding:setActionEventTextVisibility(actionEventId, false)

            _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.AXIS_MOVE_SIDE_PLAYER, self, DrivableHandtool.actionEventMoveLeftRight, false, false, true, true, nil)
            g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_VERY_HIGH)
            g_inputBinding:setActionEventTextVisibility(actionEventId, false)

            _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.AXIS_LOOK_LEFTRIGHT_PLAYER, self, DrivableHandtool.actionEventLookLeftRight, false, false, true, true, nil)
            g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_VERY_HIGH)
            g_inputBinding:setActionEventTextVisibility(actionEventId, false)
        end
    end
end

function DrivableHandtool:actionEventMoveForwardBackward(actionName, inputValue, callbackState, isAnalog)
    local spec = self.spec_drivableHandtool

    spec.inputValues.moveForward = spec.inputValues.moveForward - inputValue
end

function DrivableHandtool:actionEventMoveLeftRight(actionName, inputValue, callbackState, isAnalog)
    local spec = self.spec_drivableHandtool

    spec.inputValues.moveRight = spec.inputValues.moveRight - inputValue
end

function DrivableHandtool:actionEventLookLeftRight(actionName, inputValue, callbackState, isAnalog)
    local spec = self.spec_drivableHandtool

    inputValue = inputValue * 0.001 * g_currentDt * g_gameSettings:getValue(GameSettings.SETTING.CAMERA_SENSITIVITY)
    spec.inputValues.lookRight = spec.inputValues.lookRight + inputValue
end

----------------
---Overwrites---
----------------

---Overwritten function: getPosition
---@param superFunc function overwritten function
function DrivableHandtool:getPosition(superFunc)
    return getWorldTranslation(self.rootNode)
end

---Overwritten function: getRotation
---@param superFunc function overwritten function
function DrivableHandtool:getRotation(superFunc)
    return getWorldRotation(self.rootNode)
end

---Overwritten function: setWorldPosition
---@param superFunc function overwritten function
---@param x number world position x
---@param y number world position y
---@param z number world position z
---@param xRot number rotation x axis
---@param yRot number rotation y axis
---@param zRot number rotation z axis
---@param i integer component index
---@param changeInterp boolean change networkInterpolators
function DrivableHandtool:setWorldPosition(superFunc, x, y, z, xRot, yRot, zRot, i, changeInterp)
    superFunc(self, x, y, z, xRot, yRot, zRot, i, changeInterp)

    if self.isServer and i == 1 then
        local spec = self.spec_drivableHandtool
        local dx, _, dz = localDirectionToWorld(self.rootNode, 0, 0, 1)
        spec.currentTurnAngle = MathUtil.getYRotationFromDirection(dx, dz)

        if changeInterp then
            spec.interpolatorTurnAngle:setAngle(spec.currentTurnAngle)
        end
    end
end

---Overwritten function: setWorldPositionQuaternion
---@param superFunc function overwritten function
---@param x number world position x
---@param y number world position y
---@param z number world position z
---@param qx number quaternion roation x axis
---@param qy number quaternion roation y axis
---@param qz number quaternion roation z axis
---@param qw number quaternion roation w axis
---@param i integer component index
---@param changeInterp boolean change networkInterpolators
function DrivableHandtool:setWorldPositionQuaternion(superFunc, x, y, z, qx, qy, qz, qw, i, changeInterp)
    superFunc(self, x, y, z, qx, qy, qz, qw, i, changeInterp)

    if self.isServer and i == 1 then
        local spec = self.spec_drivableHandtool
        local dx, _, dz = localDirectionToWorld(self.rootNode, 0, 0, 1)
        spec.currentTurnAngle = MathUtil.getYRotationFromDirection(dx, dz)

        if changeInterp then
            spec.interpolatorTurnAngle:setAngle(spec.currentTurnAngle)
        end
    end
end

-- Todo: move
local function updateRotateNodeRotation(self, superFunc)
    if self.vehicle.spec_drivableHandtool ~= nil then
        self.rotY = self.origRotY
    end

    superFunc(self)
end

VehicleCamera.updateRotateNodeRotation = Utils.overwrittenFunction(VehicleCamera.updateRotateNodeRotation, updateRotateNodeRotation)
